@@ -1,5 +1,8 @@
 import re
-from collections import namedtuple, deque
+from collections import namedtuple
+
+from .utils import (_restore, split_url, decode_url, decode_url_component,
+                    remove_dot_segments)
 
 # This module based on rfc3986.
 
@@ -14,8 +17,8 @@ class InvalidPath(ValueError): pass
 class InvalidQuery(ValueError): pass
 
 
-URLTuple = namedtuple('URLBase', 'scheme host path query fragment '
-                                 'userinfo port')
+URLTuple = namedtuple('URLBase', 'scheme userinfo host port '
+                                 'path query fragment decoded')  # 4, 5, 6, 7
 
 
 class URL(URLTuple):
@@ -26,41 +29,24 @@ class URL(URLTuple):
     """
     __slots__ = ()
 
-    # This is not validating regexp.
-    # It splits url to unambiguous parts according RFC.
-    _split_re = re.compile(r'''
-        (?:([^:/?#]+):)?            # scheme
-        (?://                       # authority
-            (?:([^/?\#@]*)@)?       # userinfo
-            ([^/?\#]*)()            # host:port
-        )?
-        ([^?\#]*)                   # path
-        \??([^\#]*)                 # query
-        \#?(.*)                     # fragment
-        ''', re.VERBOSE | re.DOTALL).match
+    def __new__(cls, url=None, scheme='', userinfo='', host='', port='',
+                path='', query='', fragment=''):
 
-    def __new__(cls, url=None, scheme='', host='', path='', query='',
-                fragment='', userinfo='', port=''):
         if url is not None:
-            # All other arguments are ignored.
-            (scheme, userinfo, host, port,
-             path, query, fragment) = cls._split_re(url).groups('')
+            return cls._create_and_fix(*split_url(url))
 
-            # Host itself can contain digits and ':', so we cant use regexp.
-            # It is bit different from other behaviors: instead of splitting
-            # as is, and raise on validate, we split port only with digits.
-            # I believe this is expected behavior.
-            _port_idx = host.rfind(':')
-            if _port_idx >= 0:
-                _port = host[_port_idx + 1:]
-                if not _port or _port.isdigit():
-                    host, port = host[:_port_idx], _port
+        return cls._create_and_fix(scheme, userinfo, host, port, path,
+                                   query, fragment)
 
-        else:
-            # If url has authority, path should be absolute.
+    @classmethod
+    def _create_and_fix(cls, scheme, userinfo, host, port, path,
+                        query, fragment, decoded=False):
+        # | If a URI contains an authority component, then the path
+        # | component must either be empty or begin with a slash ("/")
+        # | character.
+        if path and path[0] != '/':
             if userinfo or host or port:
-                if path and path[0] != '/':
-                    path = '/' + path
+                path = '/' + path
 
         # | Although schemes are case-insensitive, the canonical form
         # | is lowercase. An implementation should only produce lowercase
@@ -70,19 +56,27 @@ class URL(URLTuple):
         # | should use lowercase for registered names and hexadecimal
         # | addresses for the sake of uniformity.
 
-        return tuple.__new__(cls, (scheme.lower(), host.lower(), path, query,
-                                   fragment, userinfo, str(port)))
+        return tuple.__new__(cls, (scheme.lower(), userinfo, host.lower(),
+                                   str(port), path, query, fragment, decoded))
+
+    def decode(self, encoding='utf-8', errors='replace'):
+        if self[7]:
+            return self
+        return tuple.__new__(type(self), (self[0],
+                                          decode_url(self[1], encoding, errors),
+                                          decode_url(self[2], encoding, errors),
+                                          self[3],
+                                          decode_url(self[4], encoding, errors),
+                                          decode_url(self[5], encoding, errors),
+                                          decode_url(self[6], encoding, errors),
+                                          True))
 
     ### Serialization
 
     def __unicode__(self):
-        scheme, base, path, query, fragment, userinfo, port = self
-
-        if port:
-            base += ':' + port
-
-        if userinfo:
-            base = userinfo + '@' + base
+        scheme = self[0]
+        path = self[4]
+        base = self.authority
 
         # Escape path with slashes by adding explicit empty host.
         if base or path[0:2] == '//':
@@ -97,47 +91,70 @@ class URL(URLTuple):
         if scheme:
             base = scheme + ':' + base
 
-        if query:
-            path += '?' + query
-
-        if fragment:
-            path += '#' + fragment
-
-        return base + path
+        return base + self.full_path
 
     as_string = __unicode__
 
     def __reduce__(self):
-        return type(self), (None,) + tuple(self)
+        return _restore, (type(self), tuple(self))
 
     ### Missing properties
 
+    # | The userinfo subcomponent may consist of a user name and, optionally,
+    # | scheme-specific information about how to gain authorization to access
+    # | the resource.
+    @property
+    def username(self):
+        return self[1].partition(':')[0]
+
+    @property
+    def authorization(self):
+        return self[1].partition(':')[2]
+
     @property
     def authority(self):
-        authority = self[1]
-        userinfo, port = self[5:7]
+        userinfo, base, port = self[1:4]
+
         if port:
-            authority += ':' + port
+            base += ':' + port
+
+        elif ':' in base:
+            last = base[base.rfind(':') + 1:]
+            if not last or last.isdigit():
+                base += ':'
+
         if userinfo:
-            return userinfo + '@' + authority
-        return authority
+            base = userinfo + '@' + base
+
+        elif '@' in base:
+            base = '@' + base
+
+        return base
 
     @property
     def full_path(self):
-        path, query, fragment = self[2:5]
+        path, query, fragment = self[4:7]
+
         if query:
             path += '?' + query
+
         if fragment:
             path += '#' + fragment
+
         return path
 
     ### Information
 
+    @property
+    def _data(self):
+        return self[0:7]
+
     def __nonzero__(self):
-        return any(self)
+        return any(self._data)
 
     def has_authority(self):
-        return bool(self[1] or self[5] or self[6])
+        # micro optimization: self[2] hits more often then others
+        return bool(self[2] or self[1] or self[3])
 
     def is_relative(self):
         # In terms of rfc relative url have no scheme.
@@ -145,14 +162,14 @@ class URL(URLTuple):
         return not self[0]
 
     def is_relative_path(self):
-        # Absolute path always starts with slash. Also paths with authority
-        # can not be relative.
-        return not self[2][:1] == '/' and not (
-            self[0] or self.has_authority())
+        # Relative-path url will not replace path during joining.
+        return not self[0] and (self[4][0] != '/'
+                                if self[4]
+                                else not self.has_authority())
 
     def is_host_ipv4(self):
-        if self[1][:1] != '[':
-            parts = self[1].split('.')
+        if self[2][:1] != '[':
+            parts = self[2].split('.')
             if len(parts) == 4 and all(part.isdigit() for part in parts):
                 if all(int(part, 10) < 256 for part in parts):
                     return True
@@ -162,7 +179,7 @@ class URL(URLTuple):
         if self.is_host_ipv4():
             return True
 
-        if self[1][:1] == '[' and self[1][-1:] == ']':
+        if self[2][:1] == '[' and self[2][-1:] == ']':
             return True
 
         return False
@@ -173,8 +190,6 @@ class URL(URLTuple):
     # '[' and ']' the only chars not allowed in userinfo and not delimiters
     _valid_userinfo_re = re.compile(r'^[^/?\#@\[\]]+$').match
     _valid_reg_name_re = re.compile(r'^[^/?\#@\[\]:]+$').match
-    _valid_path_re = re.compile(r'^[^?\#]+$').match
-    _valid_query_re = re.compile(r'^[^\#]+$').match
     # This primitive regular expression not match complicated ip literal.
     _valid_ip_literal_re = re.compile(r'''
         ^(?:
@@ -189,11 +204,11 @@ class URL(URLTuple):
             if not self._valid_scheme_re(self[0]):
                 raise InvalidScheme()
 
-        if self[5]:
-            if not self._valid_userinfo_re(self[5]):
+        if self[1]:
+            if not self._valid_userinfo_re(self[1]):
                 raise InvalidUserinfo()
 
-        host = self[1]
+        host = self[2]
         if host:
             if host[:1] == '[' and host[-1:] == ']':
                 if not self._valid_ip_literal_re(host[1:-1]):
@@ -207,13 +222,11 @@ class URL(URLTuple):
         # There should be no scheme and authority and first segment of path
         # should contain ':' or starts with '//'. But this library not about
         # punish user. We can escape this paths when formatting string.
-        if self[2]:
-            if not self._valid_path_re(self[2]):
-                raise InvalidPath()
+        if '?' in self[4] or '#' in self[4]:
+            raise InvalidPath()
 
-        if self[3]:
-            if not self._valid_query_re(self[3]):
-                raise InvalidQuery()
+        if '#' in self[5]:
+            raise InvalidQuery()
 
         return self
 
@@ -222,102 +235,73 @@ class URL(URLTuple):
     def __add__(self, other):
         # This method designed with one issue. By rfc delimeters without
         # following values are matter. For example:
-        # 'http://ya.ru/page' + '//?none' = 'http:'
+        # 'http://ya.ru/page' + '//?none' = 'http:?none'
         # 'http://ya.ru/page?param' + '?' = 'http://ya.ru/page'
         # But the parser makes no distinction between empty and undefined part.
         # 'http://ya.ru/page' + '//?none' = 'http://ya.ru/page?none'
         # 'http://ya.ru/page?param' + '?' = 'http://ya.ru/page?param'
         # Same bug also present in urllib.parse.urljoin.
         # I hope it will be fixed in future yurls.
-        # TODO: call remove_dot_segments() when path not modified.
 
         if not isinstance(other, URLTuple):
             raise NotImplementedError()
 
-        scheme, host, path, query, fragment, userinfo, port = other
+        scheme, userinfo, host, port, path, query, fragment = other._data
 
         if not scheme:
             scheme = self[0]
 
             if not (host or userinfo or port):
-                host, userinfo, port = self[1], self[5], self[6]
+                userinfo, host, port = self[1:4]
 
                 if not path:
-                    path = self[2]
+                    path = self[4]
 
                     if not query:
-                        query = self[3]
+                        query = self[5]
 
                 else:
                     if path[0] != '/':
-                        parts = self[2].rpartition('/')
+                        parts = self[4].rpartition('/')
                         path = parts[0] + parts[1] + path
 
-        return URL.__new__(type(self), None, scheme, host,
-                           self.remove_dot_segments(path),
-                           query, fragment, userinfo, port)
+        return self._create_and_fix(scheme, userinfo, host, port,
+                                    remove_dot_segments(path), query, fragment)
 
-    def __radd__(self, left):
-        # if other is instance of URL(), __radd__() should not be called.
-        if type(left) == URLTuple:
-            return URL.__add__(left, self)
-
-        raise NotImplementedError()
-
-    def replace(self, scheme=None, host=None, path=None, query=None,
-                fragment=None, userinfo=None, port=None, authority=None,
-                full_path=None):
+    def replace(self, scheme=None, userinfo=None, host=None, port=None,
+                path=None, query=None, fragment=None,
+                authority=None, full_path=None):
         if authority is not None:
             if host or userinfo or port:
                 raise TypeError()
 
             # Use original URL just for parse.
-            _, host, _, _, _, userinfo, port = URL('//' + authority)
+            userinfo, host, port = URL('//' + authority)[1:4]
 
         if full_path is not None:
             if path or query or fragment:
                 raise TypeError()
 
             # Use original URL just for parse.
-            path, query, fragment = URL(full_path)[2:5]
+            path, query, fragment = URL(full_path)[4:7]
 
-        return URL.__new__(type(self), None,
-                           self[0] if scheme is None else scheme,
-                           self[1] if host is None else host,
-                           self[2] if path is None else path,
-                           self[3] if query is None else query,
-                           self[4] if fragment is None else fragment,
-                           self[5] if userinfo is None else userinfo,
-                           self[6] if port is None else port)
+        return self._create_and_fix(self[0] if scheme is None else scheme,
+                                    self[1] if userinfo is None else userinfo,
+                                    self[2] if host is None else host,
+                                    self[3] if port is None else port,
+                                    self[4] if path is None else path,
+                                    self[5] if query is None else query,
+                                    self[6] if fragment is None else fragment)
 
-    def setdefault(self, scheme='', host='', path='', query='', fragment='',
-                   userinfo='', port=''):
-        return URL.__new__(type(self), None,
-                           self[0] or scheme,
-                           self[1] or host,
-                           self[2] or path,
-                           self[3] or query,
-                           self[4] or fragment,
-                           self[5] or userinfo,
-                           self[6] or port)
-
-    ### Utils
-
-    @classmethod
-    def remove_dot_segments(cls, path):
-        stack = deque()
-        last = 0
-        for segment in path.split('/'):
-            if segment == '.':
-                pass
-            elif segment == '..':
-                if len(stack):
-                    stack.pop()
-            else:
-                stack.append(segment)
-        if path.endswith(('/.', '/..')):
-            stack.append('')
-        return '/'.join(stack)
+    def setdefault(self, scheme='', userinfo='', host='', port='', path='',
+                   query='', fragment=''):
+        return self._create_and_fix(self[0] or scheme,
+                                    self[1] or userinfo,
+                                    self[2] or host,
+                                    self[3] or port,
+                                    self[4] or path,
+                                    self[5] or query,
+                                    self[6] or fragment)
 
     ### Python 2 to 3 compatibility
 
